/*                                                                           a
 *   This file is part of the ESO X-shooter Pipeline                         *
 *   Copyright (C) 2006 European Southern Observatory                        *
 *                                                                           *
 *   This library is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA    *
 *                                                                           */

/*
 * $Author: mwhite, nfernand $
 * $Date: 2023-11-07 $
 * $Revision: 0.1 $
 * $Name: not supported by cvs2svn $
 */

#include "xsh_utils_wavl_correct.h"
#include <xsh_error.h>
#include <cpl.h>
#include <cpl_error.h>
#include <math.h>
#include <hdrl_spectrum.h>
#include <hdrl_spectrum_resample.h>
#include <hdrl_image.h>
#include <hdrl_image_math.h>

#include <gsl/gsl_spline.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>

#define WAVL_OFFSET_LIM 4.0
#define WAVL_OFFSET_RMS_LIM 1.0

// Specify some global variables - this is how we are able to pass
// different spectra in to our least sq minimization functions
hdrl_spectrum1D* order_spec_leastsq;
hdrl_spectrum1D* telluric_spec_leastsq;
gsl_spline* order_spec_spline;
gsl_interp_accel* acc;

//struct fitting_data {
//    size_t n ;
//    hdrl_spectrum1D* order_spec ;
//    hdrl_spectrum1D* tell_spec ;
//};

static double get_first_valid_flux(hdrl_spectrum1D* spec) {
    int rej ;
    hdrl_value val;
    for (cpl_size i=0; i < hdrl_spectrum1D_get_size(spec); ++i) {
        val = hdrl_spectrum1D_get_flux_value(spec, i, &rej);
        cpl_msg_debug(cpl_func, "    -- Checking val(%" CPL_SIZE_FORMAT") %e, rej=%d", i, val.data, rej);
        if (rej == 0) {
            return val.data ;
        }
    }
    // Nothing found - return 0.0
    return 0.0 ;
}

static double get_last_valid_flux(hdrl_spectrum1D* spec) {
    int rej ;
    hdrl_value val;
    for (cpl_size i=hdrl_spectrum1D_get_size(spec)-1; i > -1; --i) {
        val = hdrl_spectrum1D_get_flux_value(spec, i, &rej);
        cpl_msg_debug(cpl_func, "    -- Checking val(%" CPL_SIZE_FORMAT") %e, rej=%d", i, val.data, rej);
        if (rej == 0) {
            return val.data ;
        }
    }
    // Nothing found - return 0.0
    return 0.0;
}

static void fix_spec_ends(hdrl_spectrum1D* spec, hdrl_spectrum1D* ref) {
    cpl_msg_debug(cpl_func, "    Counting spec length...");
    cpl_size len = hdrl_spectrum1D_get_size(spec);
    cpl_msg_debug(cpl_func, "    (size is %" CPL_SIZE_FORMAT ")", len);
    cpl_msg_debug(cpl_func, "    Counting spec rejected pix...");
    cpl_size bad = hdrl_image_count_rejected(spec->flux);
    if (bad == 0) {
        cpl_msg_debug(cpl_func, "    -- No bad pixels, nothing to do!");

        return ; // Nothing to do
    } else {
        cpl_msg_debug(cpl_func, "    Attempting to fix %" CPL_SIZE_FORMAT " bad spec points", bad);
    }

    // Going forward, replace any bad pixels with the first valid value
    double first = get_first_valid_flux(spec);
    first = hdrl_spectrum1D_get_flux_value(ref, 0, NULL).data;
    double last = get_last_valid_flux(spec);
    last = hdrl_spectrum1D_get_flux_value(ref,
                                          hdrl_spectrum1D_get_size(ref)-1,
                                          NULL).data;
    int rej;
    cpl_size i;
    hdrl_value val;
    // Go forward until we find a valid pix, replacing w/ first
    for (i=0; i<len; ++i) {
        val = hdrl_spectrum1D_get_flux_value(spec, i, &rej);
        if (rej != 0) {
            cpl_msg_debug(cpl_func, "    Setting pix %" CPL_SIZE_FORMAT " to %f", i, first);
            hdrl_image_accept(spec->flux, i+1, 1);
            hdrl_image_set_pixel(spec->flux, i+1, 1, (hdrl_value){first, 0.0});
        } else {
            break;
        }
    }
    // Go backward until we find a valid pix, replacing w/ last
    for (i=len-1; i>-1; --i) {
        val = hdrl_spectrum1D_get_flux_value(spec, i, &rej);
        if (rej != 0) {
            cpl_msg_debug(cpl_func, "    Setting pix %" CPL_SIZE_FORMAT "to %f", i, last);
            hdrl_image_accept(spec->flux, i+1, 1);
            hdrl_image_set_pixel(spec->flux, i+1, 1, (hdrl_value){last, 0.0});
        } else {
            break;
        }
    }
    return;
}

static hdrl_spectrum1D* basic_spectrum_interpolator(hdrl_spectrum1D* ref,
                                                    cpl_array* wavls) {
    cpl_size len = hdrl_spectrum1D_get_size(ref);
    // Create with blank flux
    hdrl_spectrum1D* to_ret = hdrl_spectrum1D_create_error_free(
            cpl_image_new(len, 1, CPL_TYPE_DOUBLE),
            wavls,
            hdrl_spectrum1D_wave_scale_linear
            );

    // Build a GSL spline to represent the reference spectrum
    acc = gsl_interp_accel_alloc();
    gsl_spline* ref_spline = gsl_spline_alloc(gsl_interp_linear,
                                              (size_t)hdrl_spectrum1D_get_size(ref));
    gsl_spline_init(ref_spline,
                    cpl_array_get_data_double(ref->wavelength),
                    cpl_image_get_data_double(hdrl_image_get_image(ref->flux)),
                    (size_t)len);

    // For each new wavelength, either:
    // - Evaluate the spline if within the original bounds; or,
    // Reject the spectrum at this location
    cpl_array* bad_pix = cpl_array_new(len, CPL_TYPE_INT);
    cpl_array_fill_window_int(bad_pix, 0, len, 0);
    double wavl_new ;
    double wavl_min = hdrl_spectrum1D_get_wavelength_value(ref, 0, NULL);
    double wavl_max = hdrl_spectrum1D_get_wavelength_value(ref, (int)len-1, NULL);
    for (cpl_size i = 0; i < len; ++i)
    {
        wavl_new = hdrl_spectrum1D_get_wavelength_value(to_ret, (int)i, NULL);
        if (wavl_new < wavl_min || wavl_new > wavl_max)
        {
            //        if (0) {
            cpl_array_set_int(bad_pix, i, 1);
        }
        else
        {
            hdrl_image_set_pixel(to_ret->flux, i + 1, 1, (hdrl_value){gsl_spline_eval(ref_spline, wavl_new, acc), 0});
        }
    }
    hdrl_spectrum1D* to_ret_masked = hdrl_spectrum1D_reject_pixels(
            to_ret, bad_pix
            );

    return to_ret_masked ;

}

static int func_f (const gsl_vector * x, void *data, gsl_vector * f) {
    cpl_msg_debug(cpl_func, "Entered func_f");

    const double p = gsl_vector_get(x, 0);

    size_t n = ((struct fitting_data *)data)->n;
    hdrl_spectrum1D* order_spec = ((struct fitting_data *)data)->order_spec;
    hdrl_spectrum1D* tell_spec = ((struct fitting_data *)data)->tell_spec;

    // Shift the order spectrum by the given amount
    hdrl_spectrum1D *order_spec_shift = hdrl_spectrum1D_duplicate(order_spec);
    cpl_msg_debug(__func__, "    Shifting order data (len %" CPL_SIZE_FORMAT ") by P = %4.3e",
                 hdrl_spectrum1D_get_size(order_spec),
                 p);
    hdrl_spectrum1D_wavelength_shift(order_spec_shift, p);


    // Resample the shifted spectrum onto the wavelength grid of the telluric
    // spectrum
//    hdrl_parameter* interp_param = hdrl_spectrum1D_resample_interpolate_parameter_create(
//            hdrl_spectrum1D_interp_akima);
    //    cpl_msg_debug(__func__, "        Interpreter formed!");
//    hdrl_spectrum1D_wavelength *tell_wav = cpl_calloc(1, sizeof(*tell_wav));
    //    cpl_msg_debug(__func__, "        Wavelength structure defined!");
    hdrl_spectrum1D_wavelength tell_wav = hdrl_spectrum1D_get_wavelength(tell_spec);
    //    cpl_msg_debug(__func__, "        Wavelengths extracted!");
    hdrl_spectrum1D* order_spec_interp;
    order_spec_interp = basic_spectrum_interpolator(order_spec_shift,
                                                    tell_wav.wavelength);
    cpl_msg_debug(cpl_func, "    Interpolated order to telluric wavls");
    cpl_msg_debug(cpl_func, "    (Interpolated spec length is %" CPL_SIZE_FORMAT ")",
                 hdrl_spectrum1D_get_size(order_spec_interp));

    // Step through and look for bad pixels - fix where possible
    cpl_msg_debug(cpl_func, "    About to try to fix spectrum ends...");
    fix_spec_ends(order_spec_interp, order_spec);
    cpl_msg_debug(cpl_func, "    Should have now corrected 'bad' pixels");
    cpl_msg_debug(cpl_func, "    (Corrected spec length is %" CPL_SIZE_FORMAT ")",
                 hdrl_spectrum1D_get_size(order_spec_interp));

    // Get the absolute difference between the spectra, and store that into the
    // input
    hdrl_spectrum1D_sub_spectrum(order_spec_interp, tell_spec);
    cpl_msg_debug(cpl_func, "    Subtracted spectra");

    // Populate the differences into the output vector
    for (size_t i=0; i < n ; ++i) {
        double diff_i = hdrl_spectrum1D_get_flux_value(order_spec_interp, (int)i, NULL).data;
        gsl_vector_set(f, i, fabs(diff_i));
    }
    cpl_msg_debug(cpl_func, "    Populated result vector");

    cleanup:
    return GSL_SUCCESS ;
}

void xsh_wc_solve_shift(gsl_vector* x0, void * params, line_analysis* analysis) {
    gsl_multifit_function_fdf fdf;
    const gsl_multifit_fdfsolver_type *T = gsl_multifit_fdfsolver_lmder ;

    const size_t n = (size_t)hdrl_spectrum1D_get_size(order_spec_leastsq);
    const size_t p = 1;
    //gsl_vector *f = gsl_vector_alloc(n);
    gsl_matrix *J = gsl_matrix_alloc(n, p);
    gsl_matrix *covar = gsl_matrix_alloc(p, p);
    // x vector already defined by calling func

    const size_t max_iter = 200;
    const double xtol = 1.0e-8 ;
    const double gtol = 1.0e-8 ;
    const double ftol = 1.0e-8 ;
    int info;

    // Define the function to be minimized
    fdf.f = func_f ;
    fdf.df = NULL ; // GSL will estimate this itself
    //fdf.fvv = NULL ; // No acceleration
    fdf.n = n ;
    fdf.p = p ;
    fdf.params = params ;

    // Allocate the workspace
    gsl_multifit_fdfsolver *work = gsl_multifit_fdfsolver_alloc(T, n, p);
    // Initialize the solver
    gsl_multifit_fdfsolver_set(work, &fdf, x0);
    // Iterate to convergence
    gsl_multifit_fdfsolver_driver(work, max_iter, xtol, gtol, ftol,
                                &info);
    // Compute cov of best fit params
    gsl_multifit_fdfsolver_jac(work,J);
    gsl_multifit_covar(J, 0.0, covar);

    gsl_vector *x = gsl_multifit_fdfsolver_position(work);
    cpl_msg_info(cpl_func, "final x = %.4e +/- %.4e\n",
                 gsl_vector_get(x, 0),
                 sqrt(gsl_matrix_get(covar, 0, 0)));
//    getchar() ;

    gsl_vector *r = gsl_multifit_fdfsolver_residual(work);

    analysis->lsf_off = gsl_vector_get(x, 0);
//    analysis->lsf_rms = sqrt(gsl_matrix_get(covar, 0, 0));
    analysis->lsf_rms = pow(gsl_blas_dnrm2(r), 1.0);
    return ;

}


static double xsh_wc_leastsq_interp_func(double x, void * p) {
    (void)(p); /* avoid unused parameter warning */
    return gsl_spline_eval(order_spec_spline, x, acc);
}

static void xsh_wc_leastsq_shift_order_spec(const double params[],
                                            hdrl_spectrum1D** order_spec) {
    cpl_msg_debug(cpl_func, "Just entered xsh_wc_leastsq_shift_order_spec");
    // Duplicate the order_spec so we can manipulate it safely
    *order_spec = hdrl_spectrum1D_duplicate(order_spec_leastsq);

    // Shift the order spectrum by the offset P specified
    cpl_msg_debug(__func__, "Shifting data by P = %4.3e", params[0]);
    check(hdrl_spectrum1D_wavelength_shift(*order_spec,
                                           -1.0 * params[0])
    );

    cleanup:

    return ;
}

static int xsh_wc_leastsq_f(const double posn[], const double params[], double* result) {
    cpl_msg_debug(cpl_func, "Just entered xsh_wc_leastsq_f");

    hdrl_spectrum1D* order_spec = cpl_calloc(1, sizeof(*order_spec)) ;
    check(xsh_wc_leastsq_shift_order_spec(params, &order_spec));

    cpl_msg_debug(__func__, "    Doing wavelength re-gridding...");
    // Re-grid the order spectrum onto the telluric wavelength scale so
    // we can directly compare at the requested wavelength value
    hdrl_parameter* interp_param = hdrl_spectrum1D_resample_interpolate_parameter_create(
            hdrl_spectrum1D_interp_akima);
    cpl_msg_debug(__func__, "        Interpreter formed!");
    hdrl_spectrum1D_wavelength *tell_wav = cpl_calloc(1, sizeof(*tell_wav));
    cpl_msg_debug(__func__, "        Wavelength structure defined!");
    check(*tell_wav = hdrl_spectrum1D_get_wavelength(telluric_spec_leastsq));
    cpl_msg_debug(__func__, "        Wavelengths extracted!");
    hdrl_spectrum1D* order_spec_interp;
    check(order_spec_interp = hdrl_spectrum1D_resample(order_spec,
                                                       tell_wav,
                                                       interp_param));

    // Clear some unused memory
    hdrl_parameter_delete(interp_param);
    hdrl_spectrum1D_delete(&order_spec);

    // Convert the requested wavelength into an array position
    cpl_array* wavl_distance = cpl_array_duplicate(order_spec_interp->wavelength);
    cpl_array_subtract_scalar(wavl_distance, posn[0]);
    cpl_array_abs(wavl_distance);
    cpl_size wavl_pos = 0;
    cpl_array_get_minpos(wavl_distance, &wavl_pos);
    cpl_array_delete(wavl_distance);

//    cpl_msg_debug(__func__, "    Setting result return for i = %f...", posn[0]);
    // Now we can compute the function value to return
    hdrl_value ret_val_hdrl;
    check(ret_val_hdrl = hdrl_spectrum1D_get_flux_value(order_spec_interp,
                                                        (int)wavl_pos,
                                                        NULL));
//    cpl_msg_debug(__func__, "    Return value computed as %f", ret_val_hdrl.data);
    check(*result = (hdrl_spectrum1D_get_flux_value(order_spec_interp,
                                                    (int)wavl_pos,
                                                    NULL).data));
//    cpl_msg_debug(__func__, "    Return result set!");
    if (isnan(*result)) {
//        cpl_msg_debug(__func__, "    Found NaN result for i=%d", (int)wavl_pos);
        double halfway = (double)hdrl_spectrum1D_get_size(order_spec_interp) / 2.0;
//        cpl_msg_debug(__func__, "    Halfway value of %f for spec length %" CPL_SIZE_FORMAT,
//                     halfway, hdrl_spectrum1D_get_size(order_spec_interp));
        int start = ((double)wavl_pos <= halfway);
        // Need to determine the first/last valid value to send back instead
        int j = (int)wavl_pos;;
        if (start) {
//            cpl_msg_debug(__func__, "start while: isnan(*result)=%d, (double)j > halfway=%d",
//                         isnan(*result), (double)j > halfway);
            while (isnan(*result) && ((double)j < halfway)) {
                ++j;
                *result = hdrl_spectrum1D_get_flux_value(order_spec_interp, j, NULL).data;
//                cpl_msg_debug(__func__, "    trying to use j=%d ==> %f", j, *result);
            }
        } else {
//            cpl_msg_debug(__func__, "end while: isnan(*result)=%d, (double)j > halfway=%d",
//                         isnan(*result), (double)j > halfway);
            while (isnan(*result) && ((double)j > halfway)) {
                --j;
                *result = hdrl_spectrum1D_get_flux_value(order_spec_interp, j, NULL).data;
//                cpl_msg_debug(__func__, "    trying to use j=%d ==> %f", j, *result);
            }
        }
    }
    cpl_msg_debug(__func__, "f = %f @ lambda = %f (pos %d)", *result, posn[0], (int)wavl_pos);

    cleanup:
//    cpl_msg_debug(__func__, "    Final memory cleanup...");
//    hdrl_spectrum1D_delete(order_spec_interp);

//    cpl_msg_debug(__func__, "    Returning success/failure code...");
//    if (cpl_error_get_code() != CPL_ERROR_NONE) {
//        cpl_msg_debug(__func__, "    ...FAIL!");
//        return 1;
//    } else {
//        cpl_msg_debug(__func__, "    ...Success!");
        return 0;
//    }

}

static int xsh_wc_leastsq_dfda(const double posn[], const double params[], double result[]) {
    cpl_msg_debug(cpl_func, "Just entered xsh_wc_leastsq_dfda");

    hdrl_spectrum1D_wavelength wavls = hdrl_spectrum1D_get_wavelength(order_spec_leastsq);
    hdrl_image* flux = hdrl_spectrum1D_get_flux(order_spec_leastsq);

    // First up - if the requested (posn - P) is outside the bounds of the input
    // spectrum, we can return 0 here
    double req_posn = posn[0] - params[0];
    if (req_posn <= cpl_array_get_min(wavls.wavelength) || req_posn >= cpl_array_get_max(wavls.wavelength)) {
        result[0] = 0.0;
        return 0;
    }


    // Now need to construct this as a GSL function so we can get a numerical
    // derivative of it
//    cpl_msg_debug(__func__, "    Setting up accelerator and spline allocation...");
//    cpl_msg_debug(__func__, "    (Attempting w/ %" CPL_SIZE_FORMAT " points)",
//                 hdrl_spectrum1D_get_size(order_spec));
    check(acc = gsl_interp_accel_alloc());
    check(order_spec_spline = gsl_spline_alloc(gsl_interp_akima,
                                               (size_t)hdrl_spectrum1D_get_size(order_spec_leastsq)));
    gsl_spline_init(order_spec_spline,
                    cpl_array_get_data_double(wavls.wavelength),
                    cpl_image_get_data_double(hdrl_image_get_image(flux)),
                    (size_t)hdrl_spectrum1D_get_size(order_spec_leastsq));
//    cpl_msg_debug(__func__, "    ...spline initialized!");

    // We now have a 'continuous' shifted spectrum. Now to differentiate it!
    // Can access a single spline value via: gsl_spline_eval(spline, xval, acc)
    gsl_function F;
    double res, abserr;

//    cpl_msg_debug(__func__, "    Preparing gsl functions...");
    F.function = &xsh_wc_leastsq_interp_func;
    F.params = 0;
    cpl_msg_debug(__func__, "interp f = %f @ lambda = %f",
                 GSL_FN_EVAL(&F, posn[0]), posn[0]);
//    cpl_msg_debug(__func__, "    Computing central derivative of GSL function...");
    double DERIV_STEP_SIZE = 1.0e-6;
    gsl_deriv_central(&F, req_posn, DERIV_STEP_SIZE, &res, &abserr);
//    cpl_msg_debug(__func__, "    ...got value %f", res);
    cpl_msg_debug(__func__, "dfda = %f @ lambda = %f", res, posn[0]);

    // Set the value in the return array
//    cpl_msg_debug(__func__, "    Building return array...");
    result[0] = res;
//    cpl_msg_debug(__func__, "    ...complete!");

    cleanup:
    gsl_spline_free(order_spec_spline);
    gsl_interp_accel_free(acc);

//    if (cpl_error_get_code == CPL_ERROR_NONE) {
        return 0;
//    } else {
//        return 1;
//    }
}

hdrl_spectrum1D* xsh_wc_molecfit_ref_to_hdrl_spectrum1D(molecfit_ref* mf) {
    cpl_image* flux;
    cpl_array* wavl;

    cpl_msg_debug(cpl_func, "Input flux min = %4.3e, Input flux max = %4.3e",
                 cpl_array_get_min(mf->telluric), cpl_array_get_max(mf->telluric));

    cpl_array* flux_intermed = cpl_array_duplicate(mf->telluric);
    cpl_array_cast(
            flux_intermed,
            CPL_TYPE_DOUBLE);
    cpl_msg_debug(cpl_func, "Intermed flux min = %4.3e, Intermed flux max = %4.3e",
                 cpl_array_get_min(flux_intermed), cpl_array_get_max(flux_intermed));

    flux = cpl_image_wrap_double(
            cpl_array_get_size(mf->telluric), 1,
            cpl_array_unwrap(cpl_array_cast(
                    cpl_array_duplicate(mf->telluric),
                    CPL_TYPE_DOUBLE)));
    wavl = cpl_array_duplicate(cpl_array_cast(mf->wavl_air, CPL_TYPE_DOUBLE));

    cpl_msg_debug(cpl_func, "Have the following flux & wavl:");
    cpl_msg_debug(cpl_func, "Flux min = %4.3e, flux max = %4.3e",
                 cpl_image_get_min(flux), cpl_image_get_max(flux));
    cpl_msg_debug(cpl_func, "Wavl min = %f, wavl max = %f",
                 cpl_array_get_min(wavl), cpl_array_get_max(wavl));

    hdrl_spectrum1D* ret_spec = hdrl_spectrum1D_create_error_free(
            cpl_image_wrap_double(
                    cpl_array_get_size(mf->telluric), 1,
                    cpl_array_unwrap(cpl_array_cast(
                            cpl_array_duplicate(mf->telluric),
                            CPL_TYPE_DOUBLE))),
            cpl_array_duplicate(cpl_array_cast(mf->wavl_air, CPL_TYPE_DOUBLE)),
            hdrl_spectrum1D_wave_scale_linear);
    return ret_spec;

}

molecfit_ref* xsh_wc_load_refdata(char* filename) {

    cpl_table* mf_table = NULL ;
    molecfit_ref* refdata = cpl_malloc(sizeof(molecfit_ref));
    cpl_msg_debug(cpl_func, ">>>PIPE-9715<<< - Have created refdata struct");

    cpl_msg_debug(cpl_func, ">>>PIPE-9715<<< - opening table %s", filename);
    mf_table = cpl_table_load(
            filename,
            1, 0);

    refdata->telluric = cpl_array_wrap_double(cpl_table_get_data_double(mf_table, "mtrans"),
                                             cpl_table_get_nrow(mf_table));
    refdata->flux = cpl_array_wrap_double(cpl_table_get_data_double(mf_table, "flux"),
                                         cpl_table_get_nrow(mf_table));

    cpl_array* wavls = cpl_array_wrap_double(cpl_table_get_data_double(mf_table, "lambda"),
                                                 cpl_table_get_nrow(mf_table));
    // Convert to nanometers
    cpl_array_multiply_scalar(wavls, 1000.0);
    cpl_msg_debug(cpl_func, "First value of input wavls: %e",
                  cpl_array_get(wavls, 0, 0));
    xsh_wc_wavl_vac_to_air(wavls);
    refdata->wavl_air = wavls;
    cpl_msg_debug(cpl_func, "First value of air wavls: %e",
                  cpl_array_get(wavls, 0, 0));

//    cpl_table_delete(mf_table);

    return refdata ;
}

int xsh_wc_wavl_vac_to_air(cpl_array* wavls) {

    cpl_msg_debug(cpl_func, "First wavl: %e", cpl_array_get(wavls, 0, 0));
    cpl_array_divide_scalar(wavls, 1000.);  // Convert nm to micron
    cpl_msg_debug(cpl_func, "First wavl after division by 1000: %e", cpl_array_get(wavls, 0, 0));

    // Define constants
    double a = 8.34213e-5;
    double b1 = 2.406030e-2;
    double c1 = 130.0;
    double b2 = 1.5997e-4;
    double c2 = 38.9;

    // The equation for this conversion is:
    // n = a + (l^2)b1/(l^2*c1 - 1) + (l^2)b2/(l^2c2 - 1) + 1
    // where l^2 is the vacuum wavelength values squared.

    // raise wavls to power 2, and make duplicates for calculation
    // This is necessary because all array-v-scalar maths in CPL is done
    // in-situ, so we need multiple pristine copies of wavls
    // Delete redundant arrays as we go
    cpl_array* wavls2 = cpl_array_duplicate(wavls);
    cpl_array_power(wavls2, 2.0);
    cpl_msg_debug(cpl_func, "First wavl after squaring: %e", cpl_array_get(wavls, 0, 0));
    cpl_array* wavls_1n = cpl_array_duplicate(wavls2);
    cpl_array* wavls_1d = cpl_array_duplicate(wavls2);
    cpl_array* wavls_2n = cpl_array_duplicate(wavls2);
    cpl_array* wavls_2d = cpl_array_duplicate(wavls2);
    cpl_array_delete(wavls2);

    // There are two array terms in the input equation - form them separately
    // eqn array term 1, numerator - wavls_1n
    cpl_array_multiply_scalar(wavls_1n, b1);
    cpl_msg_debug(cpl_func, "First wavl_1n after term: %e",
                 cpl_array_get(wavls_1n, 0, 0));
    // eqn array term 1, denominator - wavls_1d
    cpl_array_multiply_scalar(wavls_1d, c1);
    cpl_array_subtract_scalar(wavls_1d, 1.0);
    cpl_msg_debug(cpl_func, "First wavl_1d after term: %e",
                 cpl_array_get(wavls_1d, 0, 0));
    // eqn array term 2, numerator - wavls_2n
    cpl_array_multiply_scalar(wavls_2n, b2);
    cpl_msg_debug(cpl_func, "First wavl_2n after term: %e",
                 cpl_array_get(wavls_2n, 0, 0));
    // eqn array term 2, denominator - wavls_2d
    cpl_array_multiply_scalar(wavls_2d, c2);
    cpl_array_subtract_scalar(wavls_2d, 1);
    cpl_msg_debug(cpl_func, "First wavl_2d after term: %e",
                 cpl_array_get(wavls_2d, 0, 0));

    // Form the ratios, with the results in the _n arrays
    cpl_array_divide(wavls_1n, wavls_1d);
    cpl_msg_debug(cpl_func, "First wavl_1n after division: %e",
                 cpl_array_get(wavls_1n, 0, 0));
    cpl_array_divide(wavls_2n, wavls_2d);
    cpl_msg_debug(cpl_func, "First wavl_2n after division: %e",
                 cpl_array_get(wavls_2n, 0, 0));
    cpl_array_delete(wavls_1d);
    cpl_array_delete(wavls_2d);

    // Sum the terms, with the result in _1n
    cpl_array_add(wavls_1n, wavls_2n);
    cpl_msg_debug(cpl_func, "First wavl_1n after term sum: %e",
                 cpl_array_get(wavls_1n, 0, 0));
    cpl_array_delete(wavls_2n);

    // Do the rest of the equation
    cpl_array_add_scalar(wavls_1n, a + 1.0);
    cpl_msg_debug(cpl_func, "First wavl_1n after full eqn: %e",
                 cpl_array_get(wavls_1n, 0, 0));

    // Conversion equation is:
    // w_air = (w_vac/n)*1000
    cpl_array_divide(wavls, wavls_1n);
    cpl_msg_debug(cpl_func, "First wavls after completion: %e",
                 cpl_array_get(wavls, 0, 0));
    cpl_array_delete(wavls_1n);
    cpl_array_multiply_scalar(wavls, 1000.0);
    cpl_msg_debug(cpl_func, "First wavl after multiply by 1000: %e",
                 cpl_array_get(wavls, 0, 0));

    // Return 0 for success
    return 0;
}

/*
 * This function finds the local minima in a given input spectrum.
 * The input spectrum must be normalised to [0, 1].
 * If the input spectrum is a model, the local minima are absorption lines.
 * The list of lines is filtered by some criteria.
 */
cpl_array* xsh_wc_find_tell_lines(molecfit_ref* refdata) {

    cpl_size i=0;
    cpl_size j=0;
    int invalid_array_element = 1 ;
    double l = 0.0;

    // Form shifted arrays to find local minima
    cpl_array* wavl_trimmed = cpl_array_extract(refdata->wavl_air, 10, cpl_array_get_size(refdata->wavl_air) - 20);
    cpl_array* tell_trimmed = cpl_array_extract(refdata->telluric, 10, cpl_array_get_size(refdata->wavl_air) - 20);
    for (i=1 ; i<cpl_array_get_size(wavl_trimmed)-2; ++i) {
        cpl_msg_debug(cpl_func, "Check tell_trimmed value index %" CPL_SIZE_FORMAT "", i);
        check(l = cpl_array_get_double(tell_trimmed, i, NULL));
        if (!((l < cpl_array_get_double(tell_trimmed, i-1, NULL)) &&
              (l < cpl_array_get_double(tell_trimmed, i+1, NULL)))) {
            cpl_array_set_invalid(wavl_trimmed, i);
        }
    }
    cpl_array_set_invalid(wavl_trimmed, 0);
    cpl_array_set_invalid(wavl_trimmed, cpl_array_get_size(wavl_trimmed)-1);
    cpl_msg_debug(cpl_func, "Completed setting wavl_trimmed invalid posns");
    cpl_msg_debug(cpl_func, "Have flagged out %" CPL_SIZE_FORMAT " of %" CPL_SIZE_FORMAT " positions",
                 cpl_array_count_invalid(wavl_trimmed), cpl_array_get_size(wavl_trimmed));

    cpl_array* line_list = cpl_array_new(
            cpl_array_get_size(wavl_trimmed) - cpl_array_count_invalid(wavl_trimmed),
            CPL_TYPE_DOUBLE);

    j = (cpl_size)0;
    for (i=0 ; i<cpl_array_get_size(wavl_trimmed); ++i) {
        cpl_msg_debug(cpl_func, "About to ask for element %" CPL_SIZE_FORMAT " from wavl_trimmed", i);
        l = cpl_array_get_double(wavl_trimmed, i, &invalid_array_element);
        cpl_msg_debug(cpl_func,
                     "Set l = %f from wavl_trimmed posn %" CPL_SIZE_FORMAT " (invalid posn %d)",
                     l, i, invalid_array_element);
        if (invalid_array_element == 0) {
            cpl_msg_debug(cpl_func, "Setting line_list posn %" CPL_SIZE_FORMAT " from wavl_trimmed posn %" CPL_SIZE_FORMAT, j, i);
            cpl_array_set_double(line_list, j, l);
            ++j;
        }
    }

    cleanup:

    return line_list;
}

hdrl_spectrum1D* xsh_wc_get_data_order(cpl_frame* order_tab, int order) {
    cpl_table* order_table ;

    // Load the data table
    order_table = cpl_table_load(cpl_frame_get_filename(order_tab),
                                 order+1, 0);

cpl_msg_debug(cpl_func, "Trying to create data_order hdrl_spectrum1D object...");

cpl_image* flux = cpl_image_wrap_double(
        cpl_array_get_size(cpl_table_get_data_array(order_table, "FLUX1")[0]), 1,
        cpl_array_unwrap(cpl_array_cast(
                cpl_array_duplicate(cpl_table_get_data_array(order_table, "FLUX1")[0]),
                CPL_TYPE_DOUBLE)));
cpl_image* errs = cpl_image_wrap_double(
        cpl_array_get_size(cpl_table_get_data_array(order_table, "ERRS1")[0]), 1,
        cpl_array_unwrap(cpl_array_cast(
                cpl_array_duplicate(cpl_table_get_data_array(order_table, "ERRS1")[0]),
                CPL_TYPE_DOUBLE)));
cpl_array* wavls = cpl_array_duplicate(cpl_table_get_data_array(order_table, "LAMBDA")[0]);
hdrl_spectrum1D* data_order = hdrl_spectrum1D_create(flux,
           errs,
           wavls,
           hdrl_spectrum1D_wave_scale_linear);

    cleanup:

    return data_order ;
}

line_analysis* xsh_wc_analyze_line(
        molecfit_ref* telluric_data,
        hdrl_spectrum1D* order_dat,
        int order,
        double line_loc
        ) {
    cpl_size i = (cpl_size)0;
    cpl_size j = (cpl_size)0;
    double val = 0.0;
    int invalid = 0;
    cpl_bivector* bounds;

    // Create the object we will eventually return
    line_analysis* analysis = cpl_malloc(sizeof(line_analysis));

    // Create an HDRL spectrum for the telluric data
    hdrl_spectrum1D* tellspec = xsh_wc_molecfit_ref_to_hdrl_spectrum1D(telluric_data);

    // Extract the data arrays in a window around the line position
    // We can use order_data structs to neatly arrange this
    // Note - need to use cpl_array_duplicate to avoid weird
    // interactions between these arrays and their 'parents'
    bounds = cpl_bivector_new(1);
    cpl_vector_set(cpl_bivector_get_x(bounds), 0, line_loc - XSH_WAVL_CORR_GLOBAL_WD);
    cpl_vector_set(cpl_bivector_get_y(bounds), 0, line_loc + XSH_WAVL_CORR_GLOBAL_WD);

    hdrl_spectrum1D* telluric_window = hdrl_spectrum1D_select_wavelengths(
            tellspec, bounds, CPL_TRUE
            );
    hdrl_spectrum1D* order_window = hdrl_spectrum1D_select_wavelengths(
            order_dat, bounds, CPL_TRUE
            );

    cpl_msg_debug(cpl_func, "Windows for order %d, line @ %f", order, line_loc);
    cpl_msg_debug(cpl_func, "Telluric window (from %f to %f, length %" CPL_SIZE_FORMAT ", flux range %e to %e",
                 cpl_array_get_double(telluric_window->wavelength, 0, NULL),
                 cpl_array_get_double(telluric_window->wavelength,
                                      cpl_array_get_size(telluric_window->wavelength)-1, NULL),
            cpl_array_get_size(telluric_window->wavelength),
            cpl_image_get_min(hdrl_image_get_image(telluric_window->flux)),
            cpl_image_get_max(hdrl_image_get_image(telluric_window->flux))
            );
    cpl_msg_debug(cpl_func, "Order window (from %f to %f):",
                 cpl_array_get_double(order_window->wavelength, 0, NULL),
                 cpl_array_get_double(order_window->wavelength,
                                      cpl_array_get_size(order_window->wavelength)-1, NULL));
    // Normalize the flux arrays
    cpl_msg_debug(cpl_func, "Normalizing flux arrays...");
    xsh_wc_normalize_flux_array(telluric_window->flux);
    xsh_wc_normalize_flux_array(order_window->flux);
    cpl_msg_debug(cpl_func, "...complete!");

    cpl_msg_debug(cpl_func, "Telluric wavl bounds 7: %f - %f",
                 cpl_array_get_double(order_window->wavelength, 0, NULL),
                 cpl_array_get_double(order_window->wavelength, cpl_array_get_size(order_window->wavelength)-1, NULL));
    cpl_msg_debug(cpl_func, "Telluric flux bounds 7: %e - %e",
                 cpl_image_get_min(hdrl_image_get_image(order_window->flux)),
                 cpl_image_get_max(hdrl_image_get_image(order_window->flux)));
    cpl_msg_debug(cpl_func, "Telluric wavl bounds 7: %f - %f",
                 cpl_array_get_double(telluric_window->wavelength, 0, NULL),
                 cpl_array_get_double(telluric_window->wavelength, cpl_array_get_size(telluric_window->wavelength)-1, NULL));
    cpl_msg_debug(cpl_func, "Telluric flux bounds 7: %e - %e",
                 cpl_image_get_min(hdrl_image_get_image(telluric_window->flux)),
                 cpl_image_get_max(hdrl_image_get_image(telluric_window->flux)));

    // Do least-squares regression to find the best-fit shift for this line
    gsl_vector* P = gsl_vector_calloc(1);

    // find lowest *minima*
    double shift_initial = 0.0 ;
    hdrl_value curr_min = (hdrl_value){2.0, 0.0} ; // > normalized spectra
    hdrl_value curr_value ;
    for (cpl_size i=1; i < hdrl_spectrum1D_get_size(order_window)-2; ++i) {
        curr_value = hdrl_spectrum1D_get_flux_value(order_window, i, NULL);
        if (
                curr_value.data < hdrl_spectrum1D_get_flux_value(order_window, i-1, NULL).data &&
                curr_value.data < hdrl_spectrum1D_get_flux_value(order_window, i+1, NULL).data &&
                curr_value.data < curr_min.data
        ) {
            cpl_msg_debug(cpl_func, "    Found local min @ %f", (double)hdrl_spectrum1D_get_wavelength_value(
                    order_window, i, NULL));
            shift_initial = line_loc - (double)hdrl_spectrum1D_get_wavelength_value(
                    order_window, i, NULL);
            curr_min = curr_value ;
        }
    }
    // Very small, but non-zero initial shifts, cause fitter issues. If
    // the initial shift is small enough, just set it to zero.
    if (shift_initial < 0.01) {
        shift_initial = 0.0;
    }

    cpl_msg_debug(cpl_func, "Initial guess is %e", shift_initial);
    gsl_vector_set(P, 0, shift_initial);  // Initial guess

    // Build:
    // - A matrix to hold the input x coordinates;
    // - A vector to hold the input y coordinates
    cpl_matrix* wavl_vals_full = cpl_matrix_wrap(
            cpl_array_get_size(telluric_window->wavelength), 1,
            cpl_array_get_data_double(telluric_window->wavelength));
    cpl_matrix* wavl_vals = cpl_matrix_extract(wavl_vals_full,
                                               1, 0,  // Start row, column
                                               1, 1,  // Step row, column
                                               cpl_array_get_size(telluric_window->wavelength)-2, 1);
    cpl_vector* tell_flux_vals_full = cpl_vector_wrap(
            hdrl_image_get_size_x(telluric_window->flux)*hdrl_image_get_size_y(telluric_window->flux),
            cpl_image_get_data_double(hdrl_image_get_image(telluric_window->flux))
            );
    cpl_vector* tell_flux_vals = cpl_vector_extract(tell_flux_vals_full, 1, cpl_vector_get_size(tell_flux_vals_full)-2, 1);
    //cpl_msg_info(cpl_func, "Input data to LVMQ fitting:");
    //cpl_msg_info(cpl_func, "x matrix of size %" CPL_SIZE_FORMAT" x %" CPL_SIZE_FORMAT "",
     //            cpl_matrix_get_nrow(wavl_vals), cpl_matrix_get_ncol(wavl_vals));
    //cpl_msg_info(cpl_func, "y vector of size %" CPL_SIZE_FORMAT "",
    //             cpl_vector_get_size(tell_flux_vals));

    // Update the global pointers to refer to the current tell & order specs
    order_spec_leastsq = order_window;
    telluric_spec_leastsq = telluric_window;

    //cpl_msg_debug(cpl_func, "About to start LVMQ fitting...");

//    check(cpl_fit_lvmq(
//            wavl_vals, // cpl_matrix*
//            NULL, // sigma_x, must be NULL
//            tell_flux_vals, // cpl_vector*
//            NULL, // sigma_y, cpl_vector (can be NULL)
//            P, // cpl_vector
//            NULL, // int[], can be set to NULL
//            &xsh_wc_leastsq_f, // f
//            &xsh_wc_leastsq_dfda, // dfda
//            CPL_FIT_LVMQ_TOLERANCE, // relative_tolerance
//            CPL_FIT_LVMQ_COUNT, // tolerance_count
//            CPL_FIT_LVMQ_MAXITER,  // max_iterations
//            &val, // mse
//            NULL, // red_chisq
//            NULL // covariance
//            ));
    struct fitting_data data = {
            (size_t)hdrl_spectrum1D_get_size(order_window),
            order_window,
            telluric_window
    };
    xsh_wc_solve_shift(P, &data, analysis);


    cleanup:
    cpl_bivector_delete(bounds);
    hdrl_spectrum1D_delete(&tellspec);

    return analysis ;

}

cpl_error_code xsh_wc_normalize_flux_array(hdrl_image* flux) {

    cpl_image* flux_img = hdrl_image_get_image(flux);

    cpl_msg_debug(cpl_func, "    Getting flux min and max...");
    double old_min = cpl_image_get_min(flux_img);
    double old_max = cpl_image_get_max(flux_img);
    double new_min = 1;
    double new_max = 2;
    double B;
    double A;

    cpl_msg_debug(cpl_func, "    Computing components of normalization...");
    if (fabs(old_max - old_min) < 1.0e-30) {
        B = new_max - new_min;
    } else {
        B = (new_max - new_min) / (old_max - old_min);
    }
    A = ((new_max + new_min) - B * (old_max+old_min)) / 2.0 ;

    cpl_msg_debug(cpl_func, "    Applying normalization to flux array...");
    // norms = A + flux * B
    cpl_image_multiply_scalar(flux_img, B);
    cpl_image_add_scalar(flux_img, A);

    cpl_msg_debug(cpl_func, "    Normalization complete!");
    return cpl_error_get_code();
}


cpl_error_code xsh_wc_structure_output_table(cpl_table* tab) {
    check(cpl_table_new_column(tab, "ORDER", CPL_TYPE_INT));
    check(cpl_table_new_column(tab, "LINE", CPL_TYPE_DOUBLE));
    check(cpl_table_new_column(tab, "LSF_OFF", CPL_TYPE_DOUBLE));
    check(cpl_table_new_column(tab, "LSF_RMS", CPL_TYPE_DOUBLE));
    check(cpl_table_new_column(tab, "INVALID", CPL_TYPE_INT));

    cleanup:
    return cpl_error_get_code();
}

cpl_error_code xsh_wc_extract_polynomial(cpl_table* tab, int order,
                                          cpl_polynomial** poly) {

    // Need to go through the appropriate row elem-by-elem to grab out
    // the coefficients
    int is_null = 0;
    double coeff = 0.0;
    char* order_col;
    cpl_size power = 0;
    int no_coeffs = xsh_wc_get_maxdeg(order);
    *poly = cpl_polynomial_new(1);

    for (int i = 0; i <= no_coeffs; ++i) {
        order_col = cpl_sprintf("x^%d", i);
        coeff = cpl_table_get_double(tab, order_col, order-1, &is_null);
        cpl_msg_debug(cpl_func, "Have extracted %e * x^%d for order %d", coeff, i, order);
        power = (cpl_size)i;
        cpl_polynomial_set_coeff(*poly, &power, coeff);
    }

    return cpl_error_get_code();

}

cpl_array* xsh_wc_shift_wavls_poly(hdrl_spectrum1D* orig_spec,
                                   cpl_polynomial* poly) {

    cpl_msg_debug(cpl_func, "Inside xsh_wc_shift_wavls_poly");
    // Pull existing wavelength scale out
    hdrl_spectrum1D_wavelength orig_wavl = hdrl_spectrum1D_get_wavelength(orig_spec);
    cpl_msg_debug(cpl_func, "Original wavelength scale: %f - %f",
                 cpl_array_get_min(orig_wavl.wavelength),
                 cpl_array_get_max(orig_wavl.wavelength));

    // Calculate new scale
    // Owave_corr = Owave - pfinal(Owave)*scale
    cpl_array* new_wave = cpl_array_duplicate(orig_wavl.wavelength);
    for (cpl_size i=0; i < cpl_array_get_size(new_wave); ++i) {
//        shift = cpl_vector_new(1);
//        cpl_vector_set(shift, 0, cpl_array_get_double(orig_wavl.wavelength, i, NULL));
        cpl_array_set_double(
                new_wave, i,
                cpl_polynomial_eval_1d(
                        poly, cpl_array_get_double(orig_wavl.wavelength, i, NULL),
                        NULL
                )
//                cpl_polynomial_eval(poly, shift)
                );
        if (i == 0) {
            cpl_msg_debug(cpl_func, "First correction is %f @ %f",
                         cpl_array_get_double(new_wave, 0, NULL),
                         cpl_array_get_double(orig_wavl.wavelength, i, NULL));
        } else if (i + 1 == cpl_array_get_size(new_wave)) {
            cpl_msg_debug(cpl_func, "First correction is %f @ %f",
                         cpl_array_get_double(new_wave,
                                              i,
                                              NULL),
                         cpl_array_get_double(orig_wavl.wavelength, i, NULL));
            cpl_msg_debug(cpl_func, "(Should be %f)", cpl_polynomial_eval_1d(
                    poly, cpl_array_get_double(orig_wavl.wavelength, i, NULL),
                    NULL
                    ));
        }
    }
//    cpl_array_multiply_scalar(new_wave, -1.0 * XSH_WAVL_NM_PIX);
//    cpl_array_multiply_scalar(new_wave, -1.0);
    cpl_array_add(new_wave, orig_wavl.wavelength);

    cpl_msg_debug(cpl_func, "Shifted  wavelength scale: %f - %f",
                 cpl_array_get_min(new_wave),
                 cpl_array_get_max(new_wave));

    // Cleanup and return new scale
    return new_wave ;

}

cpl_error_code xsh_wc_update_res1d_frame(cpl_frame* res1D_frame,
                                         cpl_frame* res1Deso_frame,
                                         hdrl_spectrum1D* order_specs[]) {

    // Verify that the incoming file has the correct number of existing
    // extensions
    if (XSH_ORDERS_NIR != cpl_frame_get_nextensions(res1D_frame)) {
        cpl_error_set_message(
                cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                "nexts = %" CPL_SIZE_FORMAT "does not match %d",
                cpl_frame_get_nextensions(res1D_frame) - 1, XSH_ORDERS_NIR);
        return cpl_error_get_code();
    }
    cpl_msg_debug(cpl_func, "res1D_frame is %s", cpl_frame_get_filename(res1D_frame));
    if (res1Deso_frame != NULL) {
        cpl_msg_debug(cpl_func, "res1Deso_frame is %s", cpl_frame_get_filename(res1Deso_frame));
    }

    // Set up the memory we need
    cpl_propertylist* phu = NULL;
    cpl_propertylist** ext_hdrs = cpl_malloc(XSH_ORDERS_NIR * sizeof(cpl_propertylist*)) ;
    cpl_table** ext_tabs = cpl_malloc(XSH_ORDERS_NIR * sizeof(cpl_table*));
    cpl_propertylist** im_hdrs = cpl_malloc(XSH_ORDERS_NIR * sizeof(cpl_propertylist*) * 3) ;
    cpl_vector** im_data = cpl_malloc(XSH_ORDERS_NIR * sizeof(cpl_vector*) * 3);
//    const char* fn_orig = "adjusted_orig.fits";

    // Read in to memory the existing res1D_frame
    phu = cpl_propertylist_load(cpl_frame_get_filename(res1D_frame), 0);
//    cpl_propertylist_save(phu, fn_orig, CPL_IO_CREATE);
    for (int i=0; i<XSH_ORDERS_NIR; ++i) {
        ext_hdrs[i] = cpl_propertylist_load(cpl_frame_get_filename(res1D_frame), i+1);
        ext_tabs[i] = cpl_table_load(cpl_frame_get_filename(res1D_frame), i+1, 1);
        if (res1Deso_frame != NULL) {
            for (int k=0; k<3; ++k) {
                check(im_hdrs[3*i + k] = cpl_propertylist_load(cpl_frame_get_filename(res1Deso_frame), 3*i + k));
            }
        }
//        cpl_table_save(ext_tabs[i], NULL, ext_hdrs[i], fn_orig, CPL_IO_EXTEND);
    }

    // Need to update the FLUX1 column of each table
    // Note that the first and last element of the new spectra are likely
    // to be flagged invalid, so will probably need to change the QUAL1 value
    // of those as well (rather than mark the table element invalid)
    cpl_array* quals;
    cpl_array* fluxes;
    cpl_size j;
    int i = 0;
    int k = 0;
    hdrl_value flux_val = {0.0, 0.0};
    int flux_val_flag = 0;
    double* image_data_double ;
    double* image_err_double ;
    float* image_data_float ;
    float* image_err_float ;
    cpl_array* image_data_array_double ;
    cpl_array* image_err_array_double ;
    cpl_array* image_qual_array_double ;
    cpl_msg_debug(cpl_func, "Executing loop");
    for (i=0; i<XSH_ORDERS_NIR; ++i) {
        // Load up the existing QUAL and FLUX
        quals = cpl_array_duplicate(
                cpl_table_get_array(ext_tabs[i], "QUAL1", 0));
        cpl_msg_debug(cpl_func, "Max qual for ext %d = %f",
                     i, cpl_array_get_max(quals));
        cpl_msg_debug(cpl_func, "Qual array has type %s",
                     cpl_type_get_name(cpl_array_get_type(quals)));
//        getchar();
        fluxes = cpl_array_duplicate(
                cpl_table_get_array(ext_tabs[i], "FLUX1", 0));

        // Check that the array lengths all match up
        if (
                (hdrl_spectrum1D_get_size(order_specs[i]) != cpl_array_get_size(quals)) ||
                (hdrl_spectrum1D_get_size(order_specs[i]) != cpl_array_get_size(fluxes))
                ) {
            cpl_error_set_message(
                    cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                    "Spectrum size mis-match in order %d", i+1);
            return cpl_error_get_code();
        }

        // Go through the new spectrum and do the following:
        // - Mark the QUAL array when an invalid element is located
        cpl_msg_debug(cpl_func, "About to look for invalid values");
        for (j=0; j<hdrl_spectrum1D_get_size(order_specs[i]); ++j) {
            flux_val = hdrl_spectrum1D_get_flux_value(order_specs[i], j, &flux_val_flag);
            if (flux_val_flag == 1) {
                flux_val_flag = cpl_array_get_int(quals, j, NULL);
//                flux_val_flag |= (int)pow(2, 29);
                cpl_array_set_int(quals, j,
//                                  cpl_array_get_int(quals, i, NULL) & (int)pow(2, 29)
//                                  9999
                                  flux_val_flag
                                  );
            }
        }

        // Update the FLUX1, QUAL1 and ERRS1 columns of the tables
        // Note the FLUX and ERRS are wrapped up in the hdrl_image construct
        cpl_msg_debug(cpl_func, "Wrapping up new flux values...");
        image_data_double = NULL;
        cpl_msg_debug(cpl_func, "   Attempting to grab float data array...");
        check(image_data_double = cpl_image_get_data_double(hdrl_image_get_image(
                hdrl_spectrum1D_get_flux(order_specs[i])
        )));
        image_err_double = NULL;
        check(image_err_double = cpl_image_get_data_double(hdrl_image_get_error(
                hdrl_spectrum1D_get_flux(order_specs[i])
                )));
        cpl_msg_debug(cpl_func, "   Looking to wrap float data array in cpl_array...");
        image_data_float = NULL;
        image_data_float = cpl_malloc(sizeof(float_t) * hdrl_spectrum1D_get_size(order_specs[i]));
        image_err_float = NULL;
        image_err_float = cpl_malloc(sizeof(float_t) * hdrl_spectrum1D_get_size(order_specs[i]));
        for (int k=0; k<(int)hdrl_spectrum1D_get_size(order_specs[i]); ++k) {
            image_data_float[k] = (float)(image_data_double[k]);
            image_err_float[k] = (float)(image_err_double[k]);
        }
        cpl_array* image_data_array = NULL;
        cpl_array* image_err_array = NULL;
        check(image_data_array = cpl_array_wrap_float(image_data_float,
                                                      hdrl_spectrum1D_get_size(order_specs[i])));
        check(image_err_array = cpl_array_wrap_float(image_err_float,
                                                     hdrl_spectrum1D_get_size(order_specs[i])));
        check(cpl_table_set_array(ext_tabs[i], "FLUX1", 0,
                                  image_data_array));
//        check(cpl_table_set_array(ext_tabs[i], "ERRS1", 0,
//                                  image_err_array));
        cpl_msg_debug(cpl_func, "Updating QUAL1 for ext %d with max value %f",
                     i, cpl_array_get_max(quals));
        cpl_table_set_array(ext_tabs[i], "QUAL1", 0, quals);

        // Also need to update the 1D image vectors stored from the public product
        image_data_array_double = cpl_array_cast(image_data_array, CPL_TYPE_DOUBLE);
        image_err_array_double = cpl_array_cast(image_err_array, CPL_TYPE_DOUBLE);
        image_qual_array_double = cpl_array_cast(quals, CPL_TYPE_DOUBLE);
        // Need to cast all arrays to double to work with cpl_vector infrastructure
//        im_data[(int)(j*3)] = cpl_vector_duplicate(cpl_vector_wrap(cpl_array_get_size(image_data_array), image_data_float));
        cpl_msg_debug(cpl_func, "Setting up image data vector...");
        check(im_data[(int)(i*3)] = cpl_vector_wrap(cpl_array_get_size(image_data_array),
                                                    cpl_array_unwrap(image_data_array_double)));
        cpl_msg_debug(cpl_func, "Setting up image err vector...");
        check(im_data[(int)(i*3 + 1)] = cpl_vector_wrap(cpl_array_get_size(image_err_array),
                                                        cpl_array_unwrap(image_err_array_double)));
        cpl_msg_debug(cpl_func, "Setting up image qual vector...");
        check(im_data[(int)(i*3 + 2)] = cpl_vector_wrap(cpl_array_get_size(quals),
                                                        cpl_array_unwrap(image_qual_array_double)));

        // Tidy up
        cpl_array_delete(quals);
        cpl_array_delete(fluxes);
        cpl_free(image_data_float);
        cpl_free(image_err_float);

        cpl_msg_debug(cpl_func, "Completed correction loop %d", i);

//        getchar();
    }

    // Write out the results
//    const char* fn = "adjusted_order1D.fits";
    char* fn = cpl_frame_get_filename(res1D_frame);
    cpl_msg_debug(cpl_func, "Saving shifted spectra to %s", fn);
    char* im_fn = NULL;
    if (res1Deso_frame != NULL) {
        im_fn = cpl_frame_get_filename(res1Deso_frame);
        cpl_msg_debug(cpl_func, "Saving shifted spectra (ESO) to %s", im_fn);
    }
//    char* im_fn = "test_order1D.fits";
    cpl_propertylist_save(phu, fn, CPL_IO_CREATE);
    for (i=0; i<XSH_ORDERS_NIR; ++i) {
        // Internal product
        cpl_table_save(ext_tabs[i], NULL, ext_hdrs[i], fn, CPL_IO_EXTEND);
        if (res1Deso_frame != NULL) {
            // Public product
            for (k=0; k<3; ++k) {
                if (i==0 && k ==0) {
                    check(cpl_vector_save(im_data[3*i + k], im_fn,
                                          CPL_TYPE_FLOAT, im_hdrs[3*i + k], CPL_IO_CREATE));
                } else if (k == 2) {
                    check(cpl_vector_save(im_data[3*i + k], im_fn,
                                          CPL_TYPE_INT, im_hdrs[3*i + k], CPL_IO_EXTEND));
                } else {
                    check(cpl_vector_save(im_data[3*i + k], im_fn,
                                          CPL_TYPE_FLOAT, im_hdrs[3*i + k], CPL_IO_EXTEND));
                }
            }
        }
    }

    cleanup:
    cpl_msg_debug(cpl_func, "About to clear file write-out constructs...");
    for (i=0; i<XSH_ORDERS_NIR; ++i) {
        cpl_propertylist_delete(ext_hdrs[i]);
        cpl_table_delete(ext_tabs[i]);
        if (res1Deso_frame != NULL) {
            for (k=0; k<3; ++k) {
                cpl_propertylist_delete(im_hdrs[3*i + k]);
                cpl_vector_delete(im_data[3*i + k]);
            }
        }
    }
    cpl_msg_debug(cpl_func, "Clearing PHU");
    cpl_propertylist_delete(phu);
    cpl_msg_debug(cpl_func, "Clearing double pointers");
    cpl_free(ext_hdrs);
    cpl_free(ext_tabs);
    cpl_free(im_hdrs);
    cpl_free(im_data);


    return cpl_error_get_code();
}

int xsh_wc_get_maxdeg(int order) {
    if (order < XSH_WAVL_CORR_MAXDEG_SWITCH) {
        return XSH_WAVL_CORR_MAXDEG_1;
    }
    return XSH_WAVL_CORR_MAXDEG_2;
}

cpl_error_code xsh_wc_analyze_telluric(cpl_frame* mf_reference_frame,
                                       cpl_frame* res1D_frame,
                                       cpl_frame** ord_wl_offsets_tab) {

    cpl_msg_debug(cpl_func, ">>>PIPE-9715<<< - inside wavl shift calculation");

    //          get molecfic reference data
    molecfit_ref* refdata = xsh_wc_load_refdata(cpl_frame_get_filename(mf_reference_frame));
    cpl_msg_debug(cpl_func, "Telluric data *PRISTINE* has wavelength range %f to %f",
                  cpl_array_get_double(refdata->wavl_air, 0, NULL),
                  cpl_array_get_double(refdata->wavl_air, cpl_array_get_size(refdata->wavl_air) - 1, NULL));
    cpl_msg_debug(cpl_func, "Telluric data *PRISTINE* has mtrans range %f to %f",
                  cpl_array_get_double(refdata->telluric, 0, NULL),
                  cpl_array_get_double(refdata->telluric, cpl_array_get_size(refdata->telluric) - 1, NULL));
    cpl_msg_debug(cpl_func, "Telluric flux *PRISTINE* has range %4.3e to %4.3e",
                  cpl_array_get_min(refdata->flux),
                  cpl_array_get_max(refdata->flux));

    //           find the lines in the reference data
    cpl_array *tell_lines;
    check(tell_lines = xsh_wc_find_tell_lines(refdata));
    cpl_msg_debug(cpl_func, "Found %"
    CPL_SIZE_FORMAT
    " telluric lines",
            cpl_array_get_size(tell_lines));

    hdrl_spectrum1D *dataord, tellspec;

    cpl_array *tell_lines_order;
    cpl_size i = 0;
    cpl_size j = (cpl_size) 0;
    double order_wl_low = 0.0;
    double order_wl_high = 0.0;
    line_analysis *analysis;
    cpl_table *analysis_results = cpl_table_new(0);
    xsh_wc_structure_output_table(analysis_results);

    cpl_polynomial *order_fit;
    cpl_matrix *samppos;
    cpl_boolean sampsym = CPL_FALSE;
    cpl_vector *fitvals;
    cpl_vector *fitsigm;
    cpl_size maxdeg1d = 5; // Becomes 0 for order >= 7
    int good_val = 0;
    double line_val = 0.0;
    char* col_name;
    cpl_table* order_analysis_results;
    cpl_table* order_fits = cpl_table_new(XSH_ORDERS_NIR);
    cpl_table_new_column(order_fits, "ORDER", CPL_TYPE_INT);
    for (i=0; i<6; ++i) {
        col_name = cpl_sprintf("x^%" CPL_SIZE_FORMAT "", i);
        cpl_table_new_column(order_fits, col_name, CPL_TYPE_DOUBLE);
        cpl_table_fill_column_window(order_fits, col_name, 0, 16, 0.0);
    }

    cpl_boolean failed_order = CPL_FALSE;

    for (int o = 0; o < 16; o++) {

        // Grab the order spectrum (flux, wavl)
        cpl_msg_debug(cpl_func, "About to try to get flux order data from %s, of %"
        CPL_SIZE_FORMAT
        " extensions",
        cpl_frame_get_filename(res1D_frame),
        cpl_frame_get_nextensions(res1D_frame));
        cpl_frame_dump(res1D_frame, stdout);
//        getchar();
        dataord = xsh_wc_get_data_order(res1D_frame, o);

        order_wl_high = cpl_array_get_double(dataord->wavelength,
                                             cpl_array_get_size(dataord->wavelength) - 1,
                                             NULL);
        order_wl_low = cpl_array_get_double(dataord->wavelength, 0, NULL);
        cpl_msg_debug(cpl_func, "Order data %d has wavelength range %f to %f", o, order_wl_low, order_wl_high);
        cpl_msg_debug(cpl_func, "Order data %d has flux range %4.3e to %4.3e", o,
                      cpl_image_get_min(hdrl_image_get_image(dataord->flux)),
                      cpl_image_get_max(hdrl_image_get_image(dataord->flux)));
        cpl_msg_debug(cpl_func, "Telluric data has wavelength range %f to %f",
                      cpl_array_get_double(refdata->wavl_air, 0, NULL),
                      cpl_array_get_double(refdata->wavl_air, cpl_array_get_size(refdata->wavl_air) - 1, NULL));
        cpl_msg_debug(cpl_func, "Telluric flux has range %f to %f",
                      cpl_array_get_double(refdata->flux, 0, NULL),
                      cpl_array_get_double(refdata->flux, cpl_array_get_size(refdata->flux) - 1, NULL));
        cpl_msg_debug(cpl_func, "Telluric lines has wavelength range %f to %f",
                      cpl_array_get_double(tell_lines, 0, NULL),
                      cpl_array_get_double(tell_lines, cpl_array_get_size(tell_lines) - 1, NULL));

        // Identify the lines that sit inside this order
        j = (cpl_size) 0;
        tell_lines_order = cpl_array_new(cpl_array_get_size(tell_lines), CPL_TYPE_DOUBLE);
        for (i = 0; i < cpl_array_get_size(tell_lines); ++i) {
            if (((cpl_array_get_double(tell_lines, i, NULL) - XSH_WAVL_CORR_GLOBAL_WD) > order_wl_low) &&
            ((cpl_array_get_double(tell_lines, i, NULL) + XSH_WAVL_CORR_GLOBAL_WD) < order_wl_high)) {
                cpl_array_set_double(tell_lines_order, j, cpl_array_get_double(tell_lines, i, NULL));
                cpl_msg_debug(cpl_func, "Set tell_lines_order[%"
                CPL_SIZE_FORMAT
                "]=%f", j,
                cpl_array_get_double(tell_lines_order, j, NULL));
                ++j;
            }
        }

        if (cpl_array_count_invalid(tell_lines_order) < cpl_array_get_size(tell_lines_order)) {
            cpl_msg_debug(cpl_func,
                          "Have confirmed there are lines for this order, stripping line list back to valid elems");
            cpl_msg_debug(cpl_func, "Have %"
            CPL_SIZE_FORMAT
            " elems, %"
            CPL_SIZE_FORMAT
            " are invalid",
            cpl_array_get_size(tell_lines_order), cpl_array_count_invalid(tell_lines_order));
            cpl_array_set_size(tell_lines_order, cpl_array_get_size(tell_lines_order) -
            cpl_array_count_invalid(tell_lines_order));
            cpl_msg_debug(cpl_func, "Found %"
            CPL_SIZE_FORMAT
            " lines in order %d",
            cpl_array_get_size(tell_lines_order), o);

            cpl_msg_debug(cpl_func, "Have telluric data (len %"
            CPL_SIZE_FORMAT
            "):",
            cpl_array_get_size(refdata->wavl_air));
            cpl_msg_debug(cpl_func, "Telluric wavl bounds B: %f - %f",
                          cpl_array_get_double(refdata->wavl_air, 0, NULL),
                          cpl_array_get_double(refdata->wavl_air, cpl_array_get_size(refdata->wavl_air) - 1,
                                               NULL));
            cpl_msg_debug(cpl_func, "Telluric flux bounds B: %4.3e - %4.3e",
                          cpl_array_get_min(refdata->flux),
                          cpl_array_get_max(refdata->flux));

            // Analyze each of the lines for this order
            order_analysis_results = cpl_table_new(cpl_array_get_size(tell_lines_order));
            cpl_table_copy_structure(order_analysis_results, analysis_results);
            cpl_table_fill_column_window_int(order_analysis_results, "INVALID", 0,
                                             cpl_array_get_size(tell_lines_order),
                                             CPL_FALSE);
            for (i = 0; i < cpl_array_get_size(tell_lines_order); ++i) {
                // Check if the line is in range of our arrays

                cpl_msg_debug(cpl_func, "RIGHT BEFORE ANALYZE_LINE:");
                cpl_msg_debug(cpl_func, "Order data %d has wavelength range %f to %f", o, order_wl_low,
                              order_wl_high);
                cpl_msg_debug(cpl_func, "Order data %d has flux range %4.3e to %4.3e", o,
                              cpl_image_get_min(hdrl_image_get_image(dataord->flux)),
                              cpl_image_get_max(hdrl_image_get_image(dataord->flux)));
                cpl_msg_debug(cpl_func, "Analyzing line idx %"
                CPL_SIZE_FORMAT
                " in order idx %d",
                i, o);
                analysis = xsh_wc_analyze_line(refdata, dataord, o,
                                               cpl_array_get_double(tell_lines_order, i, NULL));
                cpl_table_set_int(order_analysis_results, "ORDER", i, o + 1);
                cpl_table_set_double(order_analysis_results, "LINE", i,
                                     cpl_array_get_double(tell_lines_order, i, NULL));
                cpl_table_set_double(order_analysis_results, "LSF_OFF", i, analysis->lsf_off);
                cpl_table_set_double(order_analysis_results, "LSF_RMS", i, analysis->lsf_rms);
                cpl_table_set_int(order_analysis_results, "INVALID", i, CPL_FALSE);
            }
        }

        // Trim down the table to offset and RMS limits
        for (i = 0; i < cpl_table_get_nrow(order_analysis_results); ++i) {
            if (
                    (cpl_table_get_double(order_analysis_results, "LINE", i, NULL) < 500.0) ||  // Sanity check
                    (cpl_table_get_double(order_analysis_results, "LSF_OFF", i, NULL) > WAVL_OFFSET_LIM) ||
                    (cpl_table_get_double(order_analysis_results, "LSF_OFF", i, NULL) < -1.0 * WAVL_OFFSET_LIM) ||
                    (cpl_table_get_double(order_analysis_results, "LSF_RMS", i, NULL) > WAVL_OFFSET_RMS_LIM)
                    ) {
                cpl_table_set_int(order_analysis_results, "INVALID", i, CPL_TRUE);
            }
        }

        // Append the order results to the master
        cpl_table_insert(analysis_results, order_analysis_results,
                         cpl_table_get_nrow(analysis_results));
        cpl_msg_debug(__func__, "Added results for order %d to table, now has len %"
        CPL_SIZE_FORMAT,
        o + 1, cpl_table_get_nrow(analysis_results));
        cpl_msg_debug(__func__, "Analysed %"
        CPL_SIZE_FORMAT
        " lines", cpl_array_get_size(tell_lines_order));
        cpl_msg_debug(__func__, "Contains data for orders %d to %d",
                      cpl_table_get_column_min(analysis_results, "ORDER"),
                      cpl_table_get_column_max(analysis_results, "ORDER"));
        cpl_msg_debug(__func__, "(Order-specific table contains data for orders %d to %d)",
                      cpl_table_get_column_min(order_analysis_results, "ORDER"),
                      cpl_table_get_column_max(order_analysis_results, "ORDER"));

        // Fit a polynomial to the results for this order
        for (i = 0; i < cpl_table_get_nrow(order_analysis_results); ++i) {
            if (cpl_table_get_int(order_analysis_results, "INVALID", i, NULL) == CPL_TRUE) {
                cpl_table_set_invalid(order_analysis_results, "LINE", i);
            }
        }
        cpl_msg_debug(cpl_func, "Invalidated %" CPL_SIZE_FORMAT "/%" CPL_SIZE_FORMAT " table elems for order %d",
                      cpl_table_count_invalid(order_analysis_results, "LINE"),
                      cpl_table_get_nrow(order_analysis_results),
                      o+1);

        order_fit = cpl_polynomial_new(1);
        maxdeg1d = xsh_wc_get_maxdeg(o + 1);
        //                if (o + 1 > 6) {
        //                    maxdeg1d = 0;
        //                }

        if (cpl_table_get_nrow(order_analysis_results) > cpl_table_count_invalid(order_analysis_results, "LINE")) {
            samppos = cpl_matrix_new(1, cpl_table_get_nrow(order_analysis_results) -
            cpl_table_count_invalid(order_analysis_results, "LINE"));
            fitvals = cpl_vector_new(cpl_table_get_nrow(order_analysis_results) -
                    cpl_table_count_invalid(order_analysis_results, "LINE"));
            fitsigm = cpl_vector_new(cpl_table_get_nrow(order_analysis_results) -
                    cpl_table_count_invalid(order_analysis_results, "LINE"));
            j = 0;
            for (i = 0; i < cpl_table_get_nrow(order_analysis_results); ++i) {
                line_val = cpl_table_get_double(order_analysis_results, "LINE", i, &good_val);
                if (good_val == 0) {
                    cpl_matrix_set(samppos, 0, j, line_val);
                    cpl_vector_set(fitvals, j, cpl_table_get_double(order_analysis_results, "LSF_OFF", i, NULL));
                    cpl_vector_set(fitsigm, j, cpl_table_get_double(order_analysis_results, "LSF_RMS", i, NULL));
                    ++j;
                }
            }

            cpl_polynomial_fit(order_fit, samppos, &sampsym, fitvals,
                               fitsigm, CPL_FALSE, NULL, &maxdeg1d);

            cpl_msg_debug(cpl_func, "Polynomial fitted to order %d:", o + 1);
            //                cpl_polynomial_dump(order_fit, stdout);
        } else {
            cpl_msg_warning(cpl_func, "No valid lines for fitting order %d; setting no-op polynomial",
                            o+1);
            failed_order = CPL_TRUE;
            cpl_size power = 0;
            for (i=0; i < maxdeg1d; ++i) {
                power = (cpl_size)i;
                cpl_polynomial_set_coeff(order_fit, &power, 0.0);
            }
            // getchar();
        }

        // Add order info to table
        cpl_table_set_int(order_fits, "ORDER", o, o+1);
        for (i = 0; i <= maxdeg1d; ++i) {
            col_name = cpl_sprintf("x^%" CPL_SIZE_FORMAT "", i);
            cpl_table_set_double(order_fits, col_name, o,
                                 cpl_polynomial_get_coeff(order_fit, &i));
        }


        cpl_matrix_unwrap(samppos);
        cpl_vector_unwrap(fitvals);
        cpl_vector_unwrap(fitsigm);
        cpl_table_delete(order_analysis_results);
        cpl_array_delete(tell_lines_order);
        dataord = NULL;
    }

    cpl_msg_debug(__func__, "Before final save, analysis table now has len %"
    CPL_SIZE_FORMAT,
    cpl_table_get_nrow(analysis_results));
    cpl_msg_debug(__func__, "Contains data for orders %d to %d",
                 cpl_table_get_column_min(analysis_results, "ORDER"),
                 cpl_table_get_column_max(analysis_results, "ORDER"));
    if ((cpl_msg_get_level() == CPL_MSG_DEBUG) || (cpl_msg_get_log_level() == CPL_MSG_DEBUG)) {
        cpl_propertylist* table_props = cpl_propertylist_new();
        cpl_propertylist_update_string(table_props, "ESO PRO CATG",
                                       "LINE_SHIFTS_NIR");
        cpl_table_save(analysis_results, table_props, NULL, "line_shifts.fits", CPL_IO_CREATE);
        cpl_propertylist_delete(table_props);
    }

    // Do the final output
    // PIPE-9715 - re-instate this if condition to force the system to
    // abandon wavelength fitting if an order was bad
    // if (failed_order == CPL_FALSE) {
        cpl_table_save(order_fits, NULL, NULL, "ORD_WL_OFFSETS_NIR.fits", CPL_IO_CREATE);
        *ord_wl_offsets_tab = cpl_frame_new();
        cpl_frame_set_filename(*ord_wl_offsets_tab, "ORD_WL_OFFSETS_NIR.fits");
        cpl_frame_set_tag(*ord_wl_offsets_tab, "ORD_WL_OFFSETS_NIR");
    // }

    cleanup:
    cpl_table_delete(order_fits);

    return cpl_error_get_code();
}

cpl_error_code xsh_wc_correct_for_telluric(cpl_frame* ord_wl_offsets_tab,
                                           cpl_frame** res1D_frame,
                                           cpl_frame** res1Deso_frame) {

    hdrl_spectrum1D* dataord, tellspec;

    // Open the reference data
    cpl_table* ord_polys = cpl_table_load(cpl_frame_get_filename(ord_wl_offsets_tab), 1, 0);
    cpl_msg_debug(cpl_func, "Opened table %s:", cpl_frame_get_filename(ord_wl_offsets_tab));
    //            cpl_table_dump(ord_polys, 0, 100, stdout);
    //            getchar();

    cpl_polynomial* order_poly = NULL;
    cpl_array* new_wavls = NULL;
    hdrl_parameter* regrid_params = hdrl_spectrum1D_resample_interpolate_parameter_create(
            hdrl_spectrum1D_interp_akima
//            hdrl_spectrum1D_interp_linear
            );
    hdrl_spectrum1D* order_regrid = NULL;
    hdrl_spectrum1D* order_dup = NULL ;
    hdrl_spectrum1D_wavelength dataord_wavl ;
    hdrl_spectrum1D* new_specs[XSH_ORDERS_NIR];

    // Loop over the orders and correct
    for (int o = 0; o < XSH_ORDERS_NIR; o++) {
        cpl_msg_info(cpl_func, "Correcting order %d", o+1);
        check(dataord = xsh_wc_get_data_order(*res1D_frame, o));


        // Get the polynomial from the table
        xsh_wc_extract_polynomial(ord_polys, o+1, &order_poly);
        cpl_msg_info(cpl_func, "Have extracted the following polynomial:");
        cpl_polynomial_dump(order_poly, stdout);

        // Compute the new wavelength grid
        new_wavls = xsh_wc_shift_wavls_poly(dataord, order_poly);

        // Re-grid the current order spectrum onto the new grid
        // Need to:
        // - Duplicate the existing spectrum
        // - Replace the duplicate's wavelength scale
        // - Regrid the duplicate onto the original wavl scale
        cpl_msg_debug(cpl_func, "Completed shift_wavls_poly, moving to resample...");
        cpl_msg_debug(cpl_func, "   Duplicating dataord");
        order_dup = hdrl_spectrum1D_duplicate(dataord);
        cpl_msg_debug(cpl_func, "   Re-assigning duplicate wavelength to new_wavls");
        // cpl_msg_debug(cpl_func, "      New wavls are:");
        // cpl_array_dump_structure(new_wavls, stdout);
        order_dup->wavelength = new_wavls;
        cpl_msg_debug(cpl_func, "   Grabbing wavelengths of original dataord");
        dataord_wavl = hdrl_spectrum1D_get_wavelength(dataord);
        cpl_msg_debug(cpl_func, "   Performing the resample");
        //cpl_msg_debug(cpl_func, "      At this point, what we have to play with is:");
        //cpl_msg_debug(cpl_func, "      dataord_wavl=");
        //cpl_array_dump_structure(dataord_wavl.wavelength, stdout);
        //cpl_msg_debug(cpl_func, "      order_dup->wavelength=");
        //cpl_array_dump_structure(hdrl_spectrum1D_get_wavelength(order_dup).wavelength, stdout);
        //cpl_msg_debug(cpl_func, "      order_dump->flux=");
        //hdrl_image_dump_structure(hdrl_spectrum1D_get_flux(order_dup), stdout);
        order_regrid = hdrl_spectrum1D_resample(order_dup,
                                                &dataord_wavl,
                                                regrid_params);
        cpl_msg_debug(cpl_func, "Re-gridded func back to %f - %f",
                      cpl_array_get_min(order_regrid->wavelength),
                      cpl_array_get_max(order_regrid->wavelength));
        new_specs[o] = order_regrid ;

        if ((cpl_msg_get_level() == CPL_MSG_DEBUG) || (cpl_msg_get_log_level() == CPL_MSG_DEBUG)) {
            char* fn_wavl = cpl_sprintf("wavls%02d.dat", o);
            char* fn_img_o = cpl_sprintf("flux_orig%02d.dat", o);
            char* fn_img_n = cpl_sprintf("flux_new%02d.dat", o);
            FILE* wavl_file = fopen(fn_wavl, "w");
            FILE* flux_file_orig = fopen(fn_img_o, "w");
            FILE* flux_file_new = fopen(fn_img_n, "w");
            cpl_array_dump(order_regrid->wavelength, 0, cpl_array_get_size(order_regrid->wavelength), wavl_file);
            hdrl_image_dump_window(dataord->flux, 1, 1,
                                   hdrl_image_get_size_x(dataord->flux), hdrl_image_get_size_y(dataord->flux), flux_file_orig);
            hdrl_image_dump_window(order_regrid->flux, 1, 1,
                                   hdrl_image_get_size_x(order_regrid->flux), hdrl_image_get_size_y(order_regrid->flux), flux_file_new);
            fclose(wavl_file);
            fclose(flux_file_orig);
            fclose(flux_file_new);
        }

        cpl_polynomial_delete(order_poly);
    }

    // Save back into the res1D frame
    if (res1Deso_frame != NULL) {
        check(xsh_wc_update_res1d_frame(*res1D_frame, *res1Deso_frame,
                                        new_specs));
    } else {
        check(xsh_wc_update_res1d_frame(*res1D_frame, NULL,
                                        new_specs));
    }

    cleanup:


    return cpl_error_get_code();
}